"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
This package and its contents are a work-in-progress."""

import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
from . import p4types_pb2 as _dot_p4types_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class P4Info(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PKG_INFO_FIELD_NUMBER: builtins.int
    TABLES_FIELD_NUMBER: builtins.int
    ACTIONS_FIELD_NUMBER: builtins.int
    ACTION_PROFILES_FIELD_NUMBER: builtins.int
    COUNTERS_FIELD_NUMBER: builtins.int
    DIRECT_COUNTERS_FIELD_NUMBER: builtins.int
    METERS_FIELD_NUMBER: builtins.int
    DIRECT_METERS_FIELD_NUMBER: builtins.int
    CONTROLLER_PACKET_METADATA_FIELD_NUMBER: builtins.int
    VALUE_SETS_FIELD_NUMBER: builtins.int
    REGISTERS_FIELD_NUMBER: builtins.int
    DIGESTS_FIELD_NUMBER: builtins.int
    EXTERNS_FIELD_NUMBER: builtins.int
    TYPE_INFO_FIELD_NUMBER: builtins.int
    @property
    def pkg_info(self) -> global___PkgInfo: ...
    @property
    def tables(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Table]: ...
    @property
    def actions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Action]: ...
    @property
    def action_profiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ActionProfile]: ...
    @property
    def counters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Counter]: ...
    @property
    def direct_counters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___DirectCounter]: ...
    @property
    def meters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Meter]: ...
    @property
    def direct_meters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___DirectMeter]: ...
    @property
    def controller_packet_metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControllerPacketMetadata]: ...
    @property
    def value_sets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ValueSet]: ...
    @property
    def registers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Register]: ...
    @property
    def digests(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Digest]: ...
    @property
    def externs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Extern]: ...
    @property
    def type_info(self) -> _dot_p4types_pb2.P4TypeInfo: ...
    def __init__(
        self,
        *,
        pkg_info: global___PkgInfo | None = ...,
        tables: collections.abc.Iterable[global___Table] | None = ...,
        actions: collections.abc.Iterable[global___Action] | None = ...,
        action_profiles: collections.abc.Iterable[global___ActionProfile] | None = ...,
        counters: collections.abc.Iterable[global___Counter] | None = ...,
        direct_counters: collections.abc.Iterable[global___DirectCounter] | None = ...,
        meters: collections.abc.Iterable[global___Meter] | None = ...,
        direct_meters: collections.abc.Iterable[global___DirectMeter] | None = ...,
        controller_packet_metadata: collections.abc.Iterable[global___ControllerPacketMetadata] | None = ...,
        value_sets: collections.abc.Iterable[global___ValueSet] | None = ...,
        registers: collections.abc.Iterable[global___Register] | None = ...,
        digests: collections.abc.Iterable[global___Digest] | None = ...,
        externs: collections.abc.Iterable[global___Extern] | None = ...,
        type_info: _dot_p4types_pb2.P4TypeInfo | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["pkg_info", b"pkg_info", "type_info", b"type_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["action_profiles", b"action_profiles", "actions", b"actions", "controller_packet_metadata", b"controller_packet_metadata", "counters", b"counters", "digests", b"digests", "direct_counters", b"direct_counters", "direct_meters", b"direct_meters", "externs", b"externs", "meters", b"meters", "pkg_info", b"pkg_info", "registers", b"registers", "tables", b"tables", "type_info", b"type_info", "value_sets", b"value_sets"]) -> None: ...

global___P4Info = P4Info

@typing.final
class Documentation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BRIEF_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    brief: builtins.str
    """A brief description of something, e.g. one sentence"""
    description: builtins.str
    """A more verbose description of something. Multiline is accepted. Markup
    format (if any) is TBD.
    """
    def __init__(
        self,
        *,
        brief: builtins.str = ...,
        description: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["brief", b"brief", "description", b"description"]) -> None: ...

global___Documentation = Documentation

@typing.final
class PlatformProperties(google.protobuf.message.Message):
    """Used to describe the required properties of the underlying platform."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MULTICAST_GROUP_TABLE_SIZE_FIELD_NUMBER: builtins.int
    MULTICAST_GROUP_TABLE_TOTAL_REPLICAS_FIELD_NUMBER: builtins.int
    MULTICAST_GROUP_TABLE_MAX_REPLICAS_PER_ENTRY_FIELD_NUMBER: builtins.int
    multicast_group_table_size: builtins.int
    """The minimum number of multicast entries (i.e. multicast groups) that the
    platform is required to support. If 0, there are no requirements.
    """
    multicast_group_table_total_replicas: builtins.int
    """The minimum number of replicas that the platform is required to support
    across all  groups. If 0, there are no requirements.
    """
    multicast_group_table_max_replicas_per_entry: builtins.int
    """The number of replicas that the platform is required to support per
    group/entry. If 0, `multicast_group_table_total_replicas` should be used.
    Must be no larger than `multicast_group_table_total_replicas`.
    """
    def __init__(
        self,
        *,
        multicast_group_table_size: builtins.int = ...,
        multicast_group_table_total_replicas: builtins.int = ...,
        multicast_group_table_max_replicas_per_entry: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["multicast_group_table_max_replicas_per_entry", b"multicast_group_table_max_replicas_per_entry", "multicast_group_table_size", b"multicast_group_table_size", "multicast_group_table_total_replicas", b"multicast_group_table_total_replicas"]) -> None: ...

global___PlatformProperties = PlatformProperties

@typing.final
class PkgInfo(google.protobuf.message.Message):
    """Top-level package documentation describing the forwarding pipeline config
    Can be used to manage multiple P4 packages.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    ARCH_FIELD_NUMBER: builtins.int
    ORGANIZATION_FIELD_NUMBER: builtins.int
    CONTACT_FIELD_NUMBER: builtins.int
    URL_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    PLATFORM_PROPERTIES_FIELD_NUMBER: builtins.int
    name: builtins.str
    """a definitive name for this configuration, e.g. switch.p4_v1.0"""
    version: builtins.str
    """configuration version, free-format string"""
    arch: builtins.str
    """the target architecture, e.g. "psa" """
    organization: builtins.str
    """organization which produced the configuration, e.g. "p4.org" """
    contact: builtins.str
    """contact info for support,e.g. "tech-support@acme.org" """
    url: builtins.str
    """url for more information, e.g.
    "http://support.p4.org/ref/p4/switch.p4_v1.0"
    """
    @property
    def doc(self) -> global___Documentation:
        """brief and detailed descriptions"""

    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Miscellaneous metadata, free-form; a way to extend PkgInfo"""

    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """

    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]:
        """Miscellaneous metadata, structured; a way to extend PkgInfo"""

    @property
    def platform_properties(self) -> global___PlatformProperties:
        """If set, specifies the properties that the underlying platform should have.
        If the platform does not conform to these properties, the server should
        reject the P4Info when used with a SetForwardingPipelineConfigRequest.
        """

    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
        doc: global___Documentation | None = ...,
        annotations: collections.abc.Iterable[builtins.str] | None = ...,
        annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
        arch: builtins.str = ...,
        organization: builtins.str = ...,
        contact: builtins.str = ...,
        url: builtins.str = ...,
        structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
        platform_properties: global___PlatformProperties | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["doc", b"doc", "platform_properties", b"platform_properties"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["annotation_locations", b"annotation_locations", "annotations", b"annotations", "arch", b"arch", "contact", b"contact", "doc", b"doc", "name", b"name", "organization", b"organization", "platform_properties", b"platform_properties", "structured_annotations", b"structured_annotations", "url", b"url", "version", b"version"]) -> None: ...

global___PkgInfo = PkgInfo

@typing.final
class P4Ids(google.protobuf.message.Message):
    """wrapping the enum in a message to avoid name collisions in C++, where "enum
    values are siblings of their type, not children of it"
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Prefix:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _PrefixEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[P4Ids._Prefix.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: P4Ids._Prefix.ValueType  # 0
        ACTION: P4Ids._Prefix.ValueType  # 1
        """P4 language built-ins"""
        TABLE: P4Ids._Prefix.ValueType  # 2
        VALUE_SET: P4Ids._Prefix.ValueType  # 3
        CONTROLLER_HEADER: P4Ids._Prefix.ValueType  # 4
        PSA_EXTERNS_START: P4Ids._Prefix.ValueType  # 16
        """PSA externs"""
        ACTION_PROFILE: P4Ids._Prefix.ValueType  # 17
        COUNTER: P4Ids._Prefix.ValueType  # 18
        DIRECT_COUNTER: P4Ids._Prefix.ValueType  # 19
        METER: P4Ids._Prefix.ValueType  # 20
        DIRECT_METER: P4Ids._Prefix.ValueType  # 21
        REGISTER: P4Ids._Prefix.ValueType  # 22
        DIGEST: P4Ids._Prefix.ValueType  # 23
        OTHER_EXTERNS_START: P4Ids._Prefix.ValueType  # 128
        """externs for other architectures (vendor extensions)"""
        MAX: P4Ids._Prefix.ValueType  # 255
        """max value for an unsigned 8-bit byte
        requires protoc >= 3.5.0
        reserved 0x100 to max;
        """

    class Prefix(_Prefix, metaclass=_PrefixEnumTypeWrapper):
        """ids are allocated in such a way that it is possible based on an id to
        deduce the resource type (e.g. table, action, counter, ...). The
        most-significant byte of the 32-bit id encodes the resource type. The
        purpose of this enum is to define which value is used as the
        most-significant byte for each resource type. The P4 compiler must use
        these values when allocating ids for P4 objects. Other users of P4Info can
        refer to this enum to identify a resource type based on its id.
        """

    UNSPECIFIED: P4Ids.Prefix.ValueType  # 0
    ACTION: P4Ids.Prefix.ValueType  # 1
    """P4 language built-ins"""
    TABLE: P4Ids.Prefix.ValueType  # 2
    VALUE_SET: P4Ids.Prefix.ValueType  # 3
    CONTROLLER_HEADER: P4Ids.Prefix.ValueType  # 4
    PSA_EXTERNS_START: P4Ids.Prefix.ValueType  # 16
    """PSA externs"""
    ACTION_PROFILE: P4Ids.Prefix.ValueType  # 17
    COUNTER: P4Ids.Prefix.ValueType  # 18
    DIRECT_COUNTER: P4Ids.Prefix.ValueType  # 19
    METER: P4Ids.Prefix.ValueType  # 20
    DIRECT_METER: P4Ids.Prefix.ValueType  # 21
    REGISTER: P4Ids.Prefix.ValueType  # 22
    DIGEST: P4Ids.Prefix.ValueType  # 23
    OTHER_EXTERNS_START: P4Ids.Prefix.ValueType  # 128
    """externs for other architectures (vendor extensions)"""
    MAX: P4Ids.Prefix.ValueType  # 255
    """max value for an unsigned 8-bit byte
    requires protoc >= 3.5.0
    reserved 0x100 to max;
    """

    def __init__(
        self,
    ) -> None: ...

global___P4Ids = P4Ids

@typing.final
class Preamble(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    ALIAS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    id: builtins.int
    """ids share the same number-space; e.g. table ids cannot overlap with counter
    ids. Even though this is irrelevant to this proto definition, the ids are
    allocated in such a way that it is possible based on an id to deduce the
    resource type (e.g. table, action, counter, ...). This means that code
    using these ids can detect if the wrong resource type is used
    somewhere. This also means that ids of different types can be mixed
    (e.g. direct resource list for a table) without ambiguity. Note that id 0
    is reserved and means "invalid id".
    """
    name: builtins.str
    """fully qualified name of the P4 object, e.g. c1.c2.ipv4_lpm"""
    alias: builtins.str
    """an alias (alternative name) for the P4 object, probably shorter than its
    fully qualified name. The only constraint is for it to be unique with
    respect to other P4 objects of the same type. By default, the compiler uses
    the shortest suffix of the name that uniquely identifies the object. For
    example if the P4 program contains two tables with names s.c1.t and s.c2.t,
    the default aliases will respectively be c1.t and c2.t. In the future, the
    P4 programmer may also be able to override the default alias for any P4
    object (TBD).
    """
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """

    @property
    def doc(self) -> global___Documentation:
        """Documentation of the entity"""

    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]: ...
    def __init__(
        self,
        *,
        id: builtins.int = ...,
        name: builtins.str = ...,
        alias: builtins.str = ...,
        annotations: collections.abc.Iterable[builtins.str] | None = ...,
        annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
        doc: global___Documentation | None = ...,
        structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["doc", b"doc"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["alias", b"alias", "annotation_locations", b"annotation_locations", "annotations", b"annotations", "doc", b"doc", "id", b"id", "name", b"name", "structured_annotations", b"structured_annotations"]) -> None: ...

global___Preamble = Preamble

@typing.final
class Extern(google.protobuf.message.Message):
    """used to group all extern instances of the same type in one message"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXTERN_TYPE_ID_FIELD_NUMBER: builtins.int
    EXTERN_TYPE_NAME_FIELD_NUMBER: builtins.int
    INSTANCES_FIELD_NUMBER: builtins.int
    extern_type_id: builtins.int
    """the extern_type_id is unique for a given architecture and must be in the
    range [0x81, 0xfe].
    """
    extern_type_name: builtins.str
    @property
    def instances(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ExternInstance]: ...
    def __init__(
        self,
        *,
        extern_type_id: builtins.int = ...,
        extern_type_name: builtins.str = ...,
        instances: collections.abc.Iterable[global___ExternInstance] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["extern_type_id", b"extern_type_id", "extern_type_name", b"extern_type_name", "instances", b"instances"]) -> None: ...

global___Extern = Extern

@typing.final
class ExternInstance(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    INFO_FIELD_NUMBER: builtins.int
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def info(self) -> google.protobuf.any_pb2.Any:
        """specific to the extern type, declared in a separate vendor-specific proto
        file
        """

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        info: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["info", b"info", "preamble", b"preamble"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["info", b"info", "preamble", b"preamble"]) -> None: ...

global___ExternInstance = ExternInstance

@typing.final
class MatchField(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _MatchType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _MatchTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[MatchField._MatchType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: MatchField._MatchType.ValueType  # 0
        EXACT: MatchField._MatchType.ValueType  # 2
        LPM: MatchField._MatchType.ValueType  # 3
        TERNARY: MatchField._MatchType.ValueType  # 4
        RANGE: MatchField._MatchType.ValueType  # 5
        OPTIONAL: MatchField._MatchType.ValueType  # 6

    class MatchType(_MatchType, metaclass=_MatchTypeEnumTypeWrapper): ...
    UNSPECIFIED: MatchField.MatchType.ValueType  # 0
    EXACT: MatchField.MatchType.ValueType  # 2
    LPM: MatchField.MatchType.ValueType  # 3
    TERNARY: MatchField.MatchType.ValueType  # 4
    RANGE: MatchField.MatchType.ValueType  # 5
    OPTIONAL: MatchField.MatchType.ValueType  # 6

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    BITWIDTH_FIELD_NUMBER: builtins.int
    MATCH_TYPE_FIELD_NUMBER: builtins.int
    OTHER_MATCH_TYPE_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    TYPE_NAME_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    id: builtins.int
    name: builtins.str
    bitwidth: builtins.int
    match_type: global___MatchField.MatchType.ValueType
    other_match_type: builtins.str
    """used for architecture-specific match types which are not part of the core
    P4 language or of the PSA architecture.
    """
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """

    @property
    def doc(self) -> global___Documentation:
        """Documentation of the match field"""

    @property
    def type_name(self) -> _dot_p4types_pb2.P4NamedType:
        """unset if not user-defined type"""

    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]: ...
    def __init__(
        self,
        *,
        id: builtins.int = ...,
        name: builtins.str = ...,
        annotations: collections.abc.Iterable[builtins.str] | None = ...,
        annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
        bitwidth: builtins.int = ...,
        match_type: global___MatchField.MatchType.ValueType = ...,
        other_match_type: builtins.str = ...,
        doc: global___Documentation | None = ...,
        type_name: _dot_p4types_pb2.P4NamedType | None = ...,
        structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["doc", b"doc", "match", b"match", "match_type", b"match_type", "other_match_type", b"other_match_type", "type_name", b"type_name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["annotation_locations", b"annotation_locations", "annotations", b"annotations", "bitwidth", b"bitwidth", "doc", b"doc", "id", b"id", "match", b"match", "match_type", b"match_type", "name", b"name", "other_match_type", b"other_match_type", "structured_annotations", b"structured_annotations", "type_name", b"type_name"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["match", b"match"]) -> typing.Literal["match_type", "other_match_type"] | None: ...

global___MatchField = MatchField

@typing.final
class Table(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _IdleTimeoutBehavior:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _IdleTimeoutBehaviorEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Table._IdleTimeoutBehavior.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NO_TIMEOUT: Table._IdleTimeoutBehavior.ValueType  # 0
        NOTIFY_CONTROL: Table._IdleTimeoutBehavior.ValueType  # 1

    class IdleTimeoutBehavior(_IdleTimeoutBehavior, metaclass=_IdleTimeoutBehaviorEnumTypeWrapper):
        """this enum can be extended in the future with other behaviors, such as
        "HARD_EVICTION"
        """

    NO_TIMEOUT: Table.IdleTimeoutBehavior.ValueType  # 0
    NOTIFY_CONTROL: Table.IdleTimeoutBehavior.ValueType  # 1

    PREAMBLE_FIELD_NUMBER: builtins.int
    MATCH_FIELDS_FIELD_NUMBER: builtins.int
    ACTION_REFS_FIELD_NUMBER: builtins.int
    CONST_DEFAULT_ACTION_ID_FIELD_NUMBER: builtins.int
    IMPLEMENTATION_ID_FIELD_NUMBER: builtins.int
    DIRECT_RESOURCE_IDS_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    IDLE_TIMEOUT_BEHAVIOR_FIELD_NUMBER: builtins.int
    IS_CONST_TABLE_FIELD_NUMBER: builtins.int
    HAS_INITIAL_ENTRIES_FIELD_NUMBER: builtins.int
    OTHER_PROPERTIES_FIELD_NUMBER: builtins.int
    const_default_action_id: builtins.int
    """0 (default value) means that the table does not have a const default action"""
    implementation_id: builtins.int
    """P4 id of the "implementation" for this table (e.g. action profile id); 0
    (default value) means that the table is a regular (direct) match table. As
    of today, only action profiles are supported but other table
    implementations may be added in the future
    """
    size: builtins.int
    """max number of entries in table"""
    idle_timeout_behavior: global___Table.IdleTimeoutBehavior.ValueType
    """is idle timeout supported for this table?"""
    is_const_table: builtins.bool
    """True if and only if the table's entries are immutable,
    i.e. defined using the 'const entries' table property in the P4
    source code, and thus entries cannot be deleted, modified, or
    inserted at run time.
    """
    has_initial_entries: builtins.bool
    """True if and only if the table has initial entries defined using
    the 'entries' table property in the P4 source code, either with
    or without the 'const' qualifier on 'entries', and there is at
    least one entry in that list.  This field is false if the list of
    entries is empty in the P4 source code.
    Added in 1.4.0.
    """
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def match_fields(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MatchField]: ...
    @property
    def action_refs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ActionRef]:
        """even when the table is indirect (see implementation_id) below, this field
        includes all possible actions for the table; by using ActionRef instead of
        a repeated field of action ids, each action reference in a P4 table is able
        to have its own annotations
        """

    @property
    def direct_resource_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """ids of the direct resources (if any) attached to this table; for now this
        includes only direct counters and direct meters, but other resources may be
        added in the future
        """

    @property
    def other_properties(self) -> google.protobuf.any_pb2.Any:
        """architecture-specific table properties which are not part of the core P4
        language or of the PSA architecture.
        """

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        match_fields: collections.abc.Iterable[global___MatchField] | None = ...,
        action_refs: collections.abc.Iterable[global___ActionRef] | None = ...,
        const_default_action_id: builtins.int = ...,
        implementation_id: builtins.int = ...,
        direct_resource_ids: collections.abc.Iterable[builtins.int] | None = ...,
        size: builtins.int = ...,
        idle_timeout_behavior: global___Table.IdleTimeoutBehavior.ValueType = ...,
        is_const_table: builtins.bool = ...,
        has_initial_entries: builtins.bool = ...,
        other_properties: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["other_properties", b"other_properties", "preamble", b"preamble"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["action_refs", b"action_refs", "const_default_action_id", b"const_default_action_id", "direct_resource_ids", b"direct_resource_ids", "has_initial_entries", b"has_initial_entries", "idle_timeout_behavior", b"idle_timeout_behavior", "implementation_id", b"implementation_id", "is_const_table", b"is_const_table", "match_fields", b"match_fields", "other_properties", b"other_properties", "preamble", b"preamble", "size", b"size"]) -> None: ...

global___Table = Table

@typing.final
class ActionRef(google.protobuf.message.Message):
    """used to list all possible actions in a Table"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Scope:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ScopeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ActionRef._Scope.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        TABLE_AND_DEFAULT: ActionRef._Scope.ValueType  # 0
        TABLE_ONLY: ActionRef._Scope.ValueType  # 1
        DEFAULT_ONLY: ActionRef._Scope.ValueType  # 2

    class Scope(_Scope, metaclass=_ScopeEnumTypeWrapper): ...
    TABLE_AND_DEFAULT: ActionRef.Scope.ValueType  # 0
    TABLE_ONLY: ActionRef.Scope.ValueType  # 1
    DEFAULT_ONLY: ActionRef.Scope.ValueType  # 2

    ID_FIELD_NUMBER: builtins.int
    SCOPE_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    id: builtins.int
    scope: global___ActionRef.Scope.ValueType
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """

    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]: ...
    def __init__(
        self,
        *,
        id: builtins.int = ...,
        scope: global___ActionRef.Scope.ValueType = ...,
        annotations: collections.abc.Iterable[builtins.str] | None = ...,
        annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
        structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["annotation_locations", b"annotation_locations", "annotations", b"annotations", "id", b"id", "scope", b"scope", "structured_annotations", b"structured_annotations"]) -> None: ...

global___ActionRef = ActionRef

@typing.final
class Action(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Param(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ID_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        ANNOTATIONS_FIELD_NUMBER: builtins.int
        ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
        BITWIDTH_FIELD_NUMBER: builtins.int
        DOC_FIELD_NUMBER: builtins.int
        TYPE_NAME_FIELD_NUMBER: builtins.int
        STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
        id: builtins.int
        name: builtins.str
        bitwidth: builtins.int
        @property
        def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
        @property
        def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
            """Optional. If present, the location of `annotations[i]` is given by
            `annotation_locations[i]`.
            """

        @property
        def doc(self) -> global___Documentation:
            """Documentation of the Param"""

        @property
        def type_name(self) -> _dot_p4types_pb2.P4NamedType:
            """unset if not user-defined type"""

        @property
        def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]: ...
        def __init__(
            self,
            *,
            id: builtins.int = ...,
            name: builtins.str = ...,
            annotations: collections.abc.Iterable[builtins.str] | None = ...,
            annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
            bitwidth: builtins.int = ...,
            doc: global___Documentation | None = ...,
            type_name: _dot_p4types_pb2.P4NamedType | None = ...,
            structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["doc", b"doc", "type_name", b"type_name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["annotation_locations", b"annotation_locations", "annotations", b"annotations", "bitwidth", b"bitwidth", "doc", b"doc", "id", b"id", "name", b"name", "structured_annotations", b"structured_annotations", "type_name", b"type_name"]) -> None: ...

    PREAMBLE_FIELD_NUMBER: builtins.int
    PARAMS_FIELD_NUMBER: builtins.int
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def params(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Action.Param]: ...
    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        params: collections.abc.Iterable[global___Action.Param] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["params", b"params", "preamble", b"preamble"]) -> None: ...

global___Action = Action

@typing.final
class ActionProfile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class SumOfWeights(google.protobuf.message.Message):
        """indicates that `size` and `max_group_size` represent the maximum sum of
        weights that can be present across all selector groups and within a
        single selector group respectively.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        def __init__(
            self,
        ) -> None: ...

    @typing.final
    class SumOfMembers(google.protobuf.message.Message):
        """indicates that `size` and `max_group_size` represent the maximum number
        of members that can be present across all selector groups and within a
        single selector group respectively.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        MAX_MEMBER_WEIGHT_FIELD_NUMBER: builtins.int
        max_member_weight: builtins.int
        """the maximum weight of each individual member in a group."""
        def __init__(
            self,
            *,
            max_member_weight: builtins.int | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["_max_member_weight", b"_max_member_weight", "max_member_weight", b"max_member_weight"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["_max_member_weight", b"_max_member_weight", "max_member_weight", b"max_member_weight"]) -> None: ...
        def WhichOneof(self, oneof_group: typing.Literal["_max_member_weight", b"_max_member_weight"]) -> typing.Literal["max_member_weight"] | None: ...

    PREAMBLE_FIELD_NUMBER: builtins.int
    TABLE_IDS_FIELD_NUMBER: builtins.int
    WITH_SELECTOR_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    MAX_GROUP_SIZE_FIELD_NUMBER: builtins.int
    SUM_OF_WEIGHTS_FIELD_NUMBER: builtins.int
    SUM_OF_MEMBERS_FIELD_NUMBER: builtins.int
    with_selector: builtins.bool
    """true iff the action profile used dynamic selection"""
    size: builtins.int
    """max number of member entries across all groups if the action profile does
    not have a selector. Otherwise, semantics as specified by
    `selector_size_semantics` below.
    """
    max_group_size: builtins.int
    """0 if the action profile does not have a selector. Otherwise, semantics as
    specified by `selector_size_semantics` below.
    """
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def table_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """the ids of the tables sharing this action profile"""

    @property
    def sum_of_weights(self) -> global___ActionProfile.SumOfWeights:
        """group size is the sum of the group's weights."""

    @property
    def sum_of_members(self) -> global___ActionProfile.SumOfMembers:
        """group size is the sum of the group's members."""

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        table_ids: collections.abc.Iterable[builtins.int] | None = ...,
        with_selector: builtins.bool = ...,
        size: builtins.int = ...,
        max_group_size: builtins.int = ...,
        sum_of_weights: global___ActionProfile.SumOfWeights | None = ...,
        sum_of_members: global___ActionProfile.SumOfMembers | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble", "selector_size_semantics", b"selector_size_semantics", "sum_of_members", b"sum_of_members", "sum_of_weights", b"sum_of_weights"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["max_group_size", b"max_group_size", "preamble", b"preamble", "selector_size_semantics", b"selector_size_semantics", "size", b"size", "sum_of_members", b"sum_of_members", "sum_of_weights", b"sum_of_weights", "table_ids", b"table_ids", "with_selector", b"with_selector"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["selector_size_semantics", b"selector_size_semantics"]) -> typing.Literal["sum_of_weights", "sum_of_members"] | None: ...

global___ActionProfile = ActionProfile

@typing.final
class CounterSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Unit:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _UnitEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[CounterSpec._Unit.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: CounterSpec._Unit.ValueType  # 0
        BYTES: CounterSpec._Unit.ValueType  # 1
        PACKETS: CounterSpec._Unit.ValueType  # 2
        BOTH: CounterSpec._Unit.ValueType  # 3

    class Unit(_Unit, metaclass=_UnitEnumTypeWrapper):
        """Corresponds to 'type' constructor parameter for Counter / DirectCounter in
        PSA
        """

    UNSPECIFIED: CounterSpec.Unit.ValueType  # 0
    BYTES: CounterSpec.Unit.ValueType  # 1
    PACKETS: CounterSpec.Unit.ValueType  # 2
    BOTH: CounterSpec.Unit.ValueType  # 3

    UNIT_FIELD_NUMBER: builtins.int
    unit: global___CounterSpec.Unit.ValueType
    def __init__(
        self,
        *,
        unit: global___CounterSpec.Unit.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["unit", b"unit"]) -> None: ...

global___CounterSpec = CounterSpec

@typing.final
class Counter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    INDEX_TYPE_NAME_FIELD_NUMBER: builtins.int
    size: builtins.int
    """number of entries in the counter array"""
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def spec(self) -> global___CounterSpec: ...
    @property
    def index_type_name(self) -> _dot_p4types_pb2.P4NamedType:
        """unset if index is not user-defined type"""

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        spec: global___CounterSpec | None = ...,
        size: builtins.int = ...,
        index_type_name: _dot_p4types_pb2.P4NamedType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "size", b"size", "spec", b"spec"]) -> None: ...

global___Counter = Counter

@typing.final
class DirectCounter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    DIRECT_TABLE_ID_FIELD_NUMBER: builtins.int
    direct_table_id: builtins.int
    """the id of the table to which the counter is attached"""
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def spec(self) -> global___CounterSpec: ...
    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        spec: global___CounterSpec | None = ...,
        direct_table_id: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["direct_table_id", b"direct_table_id", "preamble", b"preamble", "spec", b"spec"]) -> None: ...

global___DirectCounter = DirectCounter

@typing.final
class MeterSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Unit:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _UnitEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[MeterSpec._Unit.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: MeterSpec._Unit.ValueType  # 0
        BYTES: MeterSpec._Unit.ValueType  # 1
        PACKETS: MeterSpec._Unit.ValueType  # 2

    class Unit(_Unit, metaclass=_UnitEnumTypeWrapper):
        """Corresponds to 'type' constructor parameter for Meter / DirectMeter in PSA"""

    UNSPECIFIED: MeterSpec.Unit.ValueType  # 0
    BYTES: MeterSpec.Unit.ValueType  # 1
    PACKETS: MeterSpec.Unit.ValueType  # 2

    UNIT_FIELD_NUMBER: builtins.int
    unit: global___MeterSpec.Unit.ValueType
    def __init__(
        self,
        *,
        unit: global___MeterSpec.Unit.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["unit", b"unit"]) -> None: ...

global___MeterSpec = MeterSpec

@typing.final
class Meter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    INDEX_TYPE_NAME_FIELD_NUMBER: builtins.int
    size: builtins.int
    """number of entries in the meter array"""
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def spec(self) -> global___MeterSpec: ...
    @property
    def index_type_name(self) -> _dot_p4types_pb2.P4NamedType:
        """unset if index is not user-defined type"""

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        spec: global___MeterSpec | None = ...,
        size: builtins.int = ...,
        index_type_name: _dot_p4types_pb2.P4NamedType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "size", b"size", "spec", b"spec"]) -> None: ...

global___Meter = Meter

@typing.final
class DirectMeter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    DIRECT_TABLE_ID_FIELD_NUMBER: builtins.int
    direct_table_id: builtins.int
    """the id of the table to which the meter is attached"""
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def spec(self) -> global___MeterSpec: ...
    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        spec: global___MeterSpec | None = ...,
        direct_table_id: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["direct_table_id", b"direct_table_id", "preamble", b"preamble", "spec", b"spec"]) -> None: ...

global___DirectMeter = DirectMeter

@typing.final
class ControllerPacketMetadata(google.protobuf.message.Message):
    """Any metadata associated with controller Packet-IO (Packet-In or Packet-Out)
    is modeled as P4 headers carrying special annotations
    @controller_header("packet_out") and @controller_header("packet_in")
    respectively. There can be at most one header each with these annotations.
    This message captures the info contained within these special headers,
    and used in p4runtime.proto to supply the metadata.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Metadata(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ID_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        ANNOTATIONS_FIELD_NUMBER: builtins.int
        ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
        BITWIDTH_FIELD_NUMBER: builtins.int
        TYPE_NAME_FIELD_NUMBER: builtins.int
        STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
        id: builtins.int
        name: builtins.str
        """This is the name of the header field (not fully-qualified), similar
        to e.g. Action.Param.name.
        """
        bitwidth: builtins.int
        @property
        def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
        @property
        def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.SourceLocation]:
            """Optional. If present, the location of `annotations[i]` is given by
            `annotation_locations[i]`.
            """

        @property
        def type_name(self) -> _dot_p4types_pb2.P4NamedType:
            """unset if not user-defined type"""

        @property
        def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[_dot_p4types_pb2.StructuredAnnotation]: ...
        def __init__(
            self,
            *,
            id: builtins.int = ...,
            name: builtins.str = ...,
            annotations: collections.abc.Iterable[builtins.str] | None = ...,
            annotation_locations: collections.abc.Iterable[_dot_p4types_pb2.SourceLocation] | None = ...,
            bitwidth: builtins.int = ...,
            type_name: _dot_p4types_pb2.P4NamedType | None = ...,
            structured_annotations: collections.abc.Iterable[_dot_p4types_pb2.StructuredAnnotation] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["type_name", b"type_name"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["annotation_locations", b"annotation_locations", "annotations", b"annotations", "bitwidth", b"bitwidth", "id", b"id", "name", b"name", "structured_annotations", b"structured_annotations", "type_name", b"type_name"]) -> None: ...

    PREAMBLE_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    @property
    def preamble(self) -> global___Preamble:
        """preamble.name and preamble.id will specify header type ("packet_out" or
        "packet_in" for now).
        """

    @property
    def metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControllerPacketMetadata.Metadata]:
        """Ordered based on header layout.
        This is a constraint on the generator of this P4Info.
        """

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        metadata: collections.abc.Iterable[global___ControllerPacketMetadata.Metadata] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "preamble", b"preamble"]) -> None: ...

global___ControllerPacketMetadata = ControllerPacketMetadata

@typing.final
class ValueSet(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    MATCH_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    size: builtins.int
    """number of entries in the value_set, as per the P4 constructor call."""
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def match(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MatchField]: ...
    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        match: collections.abc.Iterable[global___MatchField] | None = ...,
        size: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["match", b"match", "preamble", b"preamble", "size", b"size"]) -> None: ...

global___ValueSet = ValueSet

@typing.final
class Register(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    TYPE_SPEC_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    INDEX_TYPE_NAME_FIELD_NUMBER: builtins.int
    size: builtins.int
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def type_spec(self) -> _dot_p4types_pb2.P4DataTypeSpec: ...
    @property
    def index_type_name(self) -> _dot_p4types_pb2.P4NamedType:
        """unset if index is not user-defined type"""

    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        type_spec: _dot_p4types_pb2.P4DataTypeSpec | None = ...,
        size: builtins.int = ...,
        index_type_name: _dot_p4types_pb2.P4NamedType | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "type_spec", b"type_spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["index_type_name", b"index_type_name", "preamble", b"preamble", "size", b"size", "type_spec", b"type_spec"]) -> None: ...

global___Register = Register

@typing.final
class Digest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PREAMBLE_FIELD_NUMBER: builtins.int
    TYPE_SPEC_FIELD_NUMBER: builtins.int
    @property
    def preamble(self) -> global___Preamble: ...
    @property
    def type_spec(self) -> _dot_p4types_pb2.P4DataTypeSpec: ...
    def __init__(
        self,
        *,
        preamble: global___Preamble | None = ...,
        type_spec: _dot_p4types_pb2.P4DataTypeSpec | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["preamble", b"preamble", "type_spec", b"type_spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["preamble", b"preamble", "type_spec", b"type_spec"]) -> None: ...

global___Digest = Digest



"Asyncio utilities."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Any, Callable

from finsy.log import LOGGER


def _create_future() -> asyncio.Future[Any]:
    "Create a new Future."
    return asyncio.get_running_loop().create_future()


async def wait_for_cancel() -> None:
    "Wait for the running task to be cancelled or interrupted."
    try:
        await _create_future()
    except BaseException as ex:
        LOGGER.debug("wait_for_cancel interrupted: %r", ex)
        raise


class CountdownFuture:
    """Implements a `Future` that we can await to count down to <= zero.

    A `CountdownFuture` can be used to track the number of running tasks and
    wait for them to finish.

    If a `CountdownFuture` is cancelled itself, it will execute a custom cancel
    function, then wait for the countdown to fully finish.
    """

    _future: asyncio.Future[int] | None = None
    _counter: int = 0

    def increment(self) -> None:
        "Increment the countdown counter."
        if self._future is None or self._future.done():
            self._future = _create_future()

        self._counter += 1

    def decrement(self) -> None:
        "Decrement the countdown counter."
        assert self._future is not None

        self._counter -= 1
        if self._counter <= 0 and not self._future.cancelled():
            self._future.set_result(1)

    def value(self) -> int:
        "Return current value of counter."
        return self._counter

    async def wait(
        self,
        on_cancel: Callable[[], None] | None = None,
    ) -> None:
        "Wait for the countdown to finish."
        if self._counter <= 0:
            return

        assert self._future is not None

        try:
            await self._future
        except asyncio.CancelledError:
            if on_cancel is not None:
                on_cancel()
            await self._wait_cancelled()
            raise

    async def _wait_cancelled(self) -> None:
        assert self._future

        while True:
            if self._counter <= 0:
                break

            self._future = _create_future()
            try:
                await self._future
            except asyncio.CancelledError:
                # Handle case where we receive subsequent cancel's but the
                # countdown still isn't finished.
                pass
            else:
                break



"P4 Value conversion functions."

# Copyright (c) 2022-2023 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network
from typing import Any, SupportsInt

from finsy.macaddr import MACAddress


class DecodeFormat(enum.Flag):
    "Indicate how to decode values."

    STRING = enum.auto()
    ADDRESS = enum.auto()
    DEFAULT = 0


def p4r_minimum_string_size(bitwidth: int) -> int:
    "P4Runtime `minimum_string_size` function (P4R-Spec section 8.4)"
    if bitwidth <= 0:
        raise ValueError(f"invalid bitwidth: {bitwidth}")
    return (bitwidth + 7) // 8


def p4r_truncate(value: bytes) -> bytes:
    "Truncate a bytes value to the specified bitwidth."
    return value.lstrip(b"\x00") or b"\x00"


def p4r_truncate_signed(value: bytes) -> bytes:
    "Truncate a signed bytes value to the specified bitwidth."
    if (value[0] & 0x80) == 0x80:
        value = value.lstrip(b"\xFF")
        if not value or (value[0] & 0x80) == 0:
            return b"\xFF" + value
        return value

    value = value.lstrip(b"\x00")
    if not value or (value[0] & 0x80) == 0x80:
        return b"\x00" + value
    return value


def all_ones(bitwidth: int) -> int:
    "Return integer value with `bitwidth` 1's."
    return (1 << bitwidth) - 1


def mask_to_prefix(value: int, bitwidth: int) -> int:
    "Convert mask bits to prefix count. Return -1 if mask is discontiguous."
    mask = ~value & all_ones(bitwidth)  # complement the bits
    if (mask & (mask + 1)) != 0:
        return -1  # discontiguous
    return bitwidth - mask.bit_length()


def _invalid_err(kind: str, bitwidth: int, value: Any) -> ValueError:
    "Construct formatted ValueError message."
    # Replace "exact" with empty string. Otherwise, prepend a space.
    if kind == "exact":
        kind = ""
    else:
        kind = f" {kind.upper()}"
    return ValueError(f"invalid{kind} value for bitwidth {bitwidth}: {value!r}")


def _parse_exact_str(value: str, bitwidth: int) -> int:
    "Convert string value to an integer."
    value = value.strip()

    if bitwidth == 32 and "." in value:
        return int(IPv4Address(value))

    if bitwidth == 128 and ":" in value:
        return int(IPv6Address(value))

    if bitwidth == 48 and ("-" in value[1:] or ":" in value):
        return int(MACAddress(value))

    try:
        return int(value, base=0)
    except ValueError:
        raise _invalid_err("exact", bitwidth, value) from None


def _decode_addr(value: int, bitwidth: int, format: DecodeFormat):
    "Decode an integer according to the bitwidth and desired output format."
    assert format & DecodeFormat.ADDRESS

    match bitwidth:
        case 128:
            addr = IPv6Address(value)
        case 48:
            addr = MACAddress(value)
        case 32:
            addr = IPv4Address(value)
        case _:
            # Return original integer.
            if format & DecodeFormat.STRING:
                return hex(value)
            return value

    if format & DecodeFormat.STRING:
        return str(addr)

    return addr


# Exact Values
# ~~~~~~~~~~~~
#
# Supported input values for `encode_exact`:
#
# - int
# - str
#   o decimal integer
#   o hexadecimal integer (prefixed with `0x`)
#   o IPv4 string (bitwidth=32 only)
#   o IPv6 string (bitwidth=128 only)
#   o MAC string  (bitwidth=48 only)
#   o SdnString (bitwidth=0)
# - IPv4Address (bitwidth=32 only)
# - IPv6Address (bitwidth=128 only)
# - MACAddress (bitwidth=48 only)
#
# (To pass `bytes``, first convert to an int using `int.from_bytes`.)
#
# Canonical output values from `decode_exact`:
#
# - DEFAULT:
#   o int
# - STRING:
#   o hexadecimal integer prefixed with `0x`
#   o SdnString (bitwidth=0)
# - ADDRESS:
#   o IPv4Address (bitwidth=32)
#   o IPv6Address (bitwidth=128)
#   o MACAddress (bitwidth=48)
#   o see DEFAULT
# - ADDRESS, STRING:
#   o IPv4 string (bitwidth=32)
#   o IPv6 string (bitwidth=128)
#   o MAC string (bitwidth=48)
#   o hexadecimal integer (all other bitwidths)

_ExactValue = int | str | IPv4Address | IPv6Address | MACAddress
_ExactReturn = int | str | IPv4Address | IPv6Address | MACAddress

P4ParamValue = _ExactValue


def encode_exact(value: _ExactValue, bitwidth: int, mask: int = ~0) -> bytes:
    """Encode an exact field value into a byte encoding.

    If `bitwidth` is 0, value is an `SdnString`.
    """
    assert value is not None

    if bitwidth == 0:
        # If bitwidth is 0, value is interpreted as `SdnString`.
        if not isinstance(value, str):
            raise ValueError(f"invalid value for SdnString: {value!r}")
        return value.encode()

    match value:
        case int():
            ival = value
        case str():
            ival = _parse_exact_str(value, bitwidth)
        case IPv4Address() if bitwidth == 32:
            ival = int(value)
        case IPv6Address() if bitwidth == 128:
            ival = int(value)
        case MACAddress() if bitwidth == 48:
            ival = int(value)
        case _:
            raise _invalid_err("exact", bitwidth, value)

    if ival >= (1 << bitwidth) or ival < 0:
        raise _invalid_err("exact", bitwidth, value)

    size = p4r_minimum_string_size(bitwidth)
    if mask != ~0:
        ival &= mask
    return p4r_truncate(ival.to_bytes(size, "big"))


def decode_exact(
    data: bytes,
    bitwidth: int,
    format: DecodeFormat = DecodeFormat.DEFAULT,
) -> _ExactReturn:
    """Decode a P4R value into an integer or address."""
    if bitwidth == 0:
        # If bitwidth is 0, data is interpreted as `SdnString`.
        return data.decode()

    if not data:
        raise _invalid_err("exact", bitwidth, data)

    ival = int.from_bytes(data, "big")
    if ival >= (1 << bitwidth):
        raise _invalid_err("exact", bitwidth, data)

    if format & DecodeFormat.ADDRESS:
        return _decode_addr(ival, bitwidth, format)

    if format & DecodeFormat.STRING:
        return hex(ival)

    return ival


def format_exact(value: _ExactValue, bitwidth: int, format: DecodeFormat) -> str:
    "Format a value as a string."
    data = encode_exact(value, bitwidth)
    result = decode_exact(data, bitwidth, format | DecodeFormat.STRING)
    assert isinstance(result, str)
    return result


# LPM Values
# ~~~~~~~~~~
#
# Supported input values for `encode_lpm`:
#
# - int
# - str
#   o decimal integer
#   o hexadecimal integer (prefixed with `0x`)
#   o IPv4 string (bitwidth=32 only)
#   o IPv6 string (bitwidth=128 only)
#   o MAC string  (bitwidth=48 only)
#   o "value/prefix" where value is any of the above
# - IPv4Address (bitwidth=32 only)
# - IPv6Address (bitwidth=128 only)
# - MACAddress (bitwidth=48 only)
# - IPv4Network (bitwidth=32 only)
# - IPv6Network (bitwidth=128 only)
# - tuple[value, prefix]
#   o tuple[int, int]
#   o tuple[str, int]
#   o tuple[IPv4Address, int] (bitwidth=32 only)
#   o tuple[IPv6Address, int] (bitwidth=128 only)
#   o tuple[MACAddress, int]  (bitwidth=48 only)
#
# Supported output values for `decode_lpm`:
#
# - DEFAULT:
#   o tuple[int, int]
# - STRING:
#   o "value/prefix" with value in hexadecimal
# - ADDRESS:
#   o IPv4Network (bitwidth=32)
#   o IPv6Network (bitwidth=128)
#   o tuple[MACAddress, int] (bitwidth=48)
#   o See DEFAULT output.
# - ADDRESS, STRING
#   o "IPV4/prefix" (bitwidth=32)
#   o "IPV6/prefix" (bitwidth=128)
#   o "MAC/prefix"  (bitwidth=48)
#   o See STRING output.

_LPMValue = (
    int
    | str
    | IPv4Network
    | IPv6Network
    | IPv4Address
    | IPv6Address
    | MACAddress
    | tuple[_ExactValue, int]
)
_LPMReturn = str | IPv4Network | IPv6Network | tuple[int | MACAddress, int]


def _parse_lpm_prefix(value: str, bitwidth: int) -> int:
    "Parse LPM prefix."
    if bitwidth == 32 and "." in value:
        return mask_to_prefix(int(IPv4Address(value)), 32)
    if bitwidth == 128 and ":" in value:
        return mask_to_prefix(int(IPv6Address(value)), 128)
    if bitwidth == 48 and ("-" in value[1:] or ":" in value):
        return mask_to_prefix(int(MACAddress(value)), 48)
    return int(value)


def _parse_lpm_str(value: str, bitwidth: int) -> tuple[bytes, int]:
    "Parse value in slash notation."
    if "/" not in value:
        return (encode_exact(value, bitwidth), bitwidth)

    vals = value.split("/", 1)
    prefix = _parse_lpm_prefix(vals[1], bitwidth)

    if prefix > bitwidth or prefix < 0:
        raise _invalid_err("lpm", bitwidth, value)

    mask = ~all_ones(bitwidth - prefix)
    return (encode_exact(vals[0], bitwidth, mask), prefix)


def encode_lpm(value: _LPMValue, bitwidth: int) -> tuple[bytes, int]:
    "Encode a value into a P4R LPM value."
    assert value is not None

    if bitwidth == 0:
        raise _invalid_err("lpm", bitwidth, value)

    match value:
        case int():
            return (encode_exact(value, bitwidth), bitwidth)
        case str():
            return _parse_lpm_str(value, bitwidth)
        case (val, int(prefix)):
            if prefix > bitwidth or prefix < 0:
                raise _invalid_err("lpm", bitwidth, value)
            mask = ~all_ones(bitwidth - prefix)
            return (encode_exact(val, bitwidth, mask), prefix)
        case IPv4Network() | IPv6Network():
            if bitwidth != value.max_prefixlen:
                raise _invalid_err("lpm", bitwidth, value)
            return (encode_exact(value.network_address, bitwidth), value.prefixlen)
        case IPv4Address() | IPv6Address() | MACAddress():
            if bitwidth != value.max_prefixlen:
                raise _invalid_err("lpm", bitwidth, value)
            return (encode_exact(value, bitwidth), bitwidth)
        case _:  # pyright: ignore[reportUnnecessaryComparison]
            raise _invalid_err("lpm", bitwidth, value)


def decode_lpm(
    data: bytes,
    prefix_len: int,
    bitwidth: int,
    format: DecodeFormat = DecodeFormat.DEFAULT,
) -> _LPMReturn:
    "Decode a P4R Value into an integer or address."
    value = decode_exact(data, bitwidth, format)

    match value:
        case IPv4Address():
            return IPv4Network((value, prefix_len))
        case IPv6Address():
            return IPv6Network((value, prefix_len))
        case str():
            if prefix_len == bitwidth:
                return value
            return f"{value}/{prefix_len}"
        case _:
            return (value, prefix_len)


def format_lpm(value: _LPMValue, bitwidth: int, format: DecodeFormat) -> str:
    "Format a value as a string."
    data = encode_lpm(value, bitwidth)
    result = decode_lpm(data[0], data[1], bitwidth, format | DecodeFormat.STRING)
    assert isinstance(result, str)
    return result


# Ternary Values
# ~~~~~~~~~~~~~~
#
# Supported input values for `encode_ternary`:
#
# - int
# - str
#   o decimal integer [exact]
#   o hexadecimal integer (prefixed with `0x`) [exact]
#   o IPv4 string (bitwidth=32 only) [exact]
#   o IPv6 string (bitwidth=128 only) [exact]
#   o MAC string  (bitwidth=48 only) [exact]
#   o "value/prefix" where value is any of the above singular values
#   o "value/&mask" where value/mask is any of the above singular values
# - IPv4Address (bitwidth=32 only) [exact]
# - IPv6Address (bitwidth=128 only) [exact]
# - MACAddress (bitwidth=48 only) [exact]
# - IPv4Network (bitwidth=32 only)
# - IPv6Network (bitwidth=128 only)
# - tuple[value, mask]
#   o tuple[int, int]
#   o tuple[str, str]
#   o tuple[IPv4Address, IPv4Address] (bitwidth=32 only)
#   o tuple[IPv6Address, IPv6Address] (bitwidth=128 only)
#   o tuple[MACAddress, MACAddress]  (bitwidth=48 only)
#
# Supported output values for `decode_ternary`:
#
# - DEFAULT:
#   o tuple[int, int]
# - STRING:
#   o "value/&mask" with value, mask in hexadecimal
# - ADDRESS:
#   o tuple[IPv4Address, IPv4Address] (bitwidth=32)
#   o tuple[IPv6Address, IPv6Address] (bitwidth=128)
#   o tuple[MACAddress, MACAddress] (bitwidth=48)
#   o See DEFAULT output.
# - ADDRESS, STRING
#   o "IPV4/&mask"  (bitwidth=32)
#   o "IPV6/&mask"  (bitwidth=128)
#   o "MAC/&mask"   (bitwidth=48)
#   o See STRING output.

_TernaryValue = (
    SupportsInt | str | IPv4Network | IPv6Network | tuple[_ExactValue, _ExactValue]
)
_TernaryReturn = tuple[_ExactReturn, _ExactReturn] | _ExactReturn


def _parse_ternary_str(value: str, bitwidth: int) -> tuple[bytes, bytes]:
    "Parse value in slash or slash-amp notation."
    if "/&" not in value:
        data, prefix = _parse_lpm_str(value, bitwidth)
        mask = all_ones(prefix) << (bitwidth - prefix)
        return (data, encode_exact(mask, bitwidth))

    vals = value.split("/&", 1)

    # TODO: Masked bits in value must be zero.
    return (encode_exact(vals[0], bitwidth), encode_exact(vals[1], bitwidth))


def encode_ternary(value: _TernaryValue, bitwidth: int) -> tuple[bytes, bytes]:
    "Encode a value into a P4R ternary value."
    assert value is not None

    match value:
        case int(val):
            mask = all_ones(bitwidth)
        case str():
            return _parse_ternary_str(value, bitwidth)
        case (val, mask):
            pass
        case IPv4Network() | IPv6Network():
            val, mask = int(value.network_address), int(value.netmask)
        case IPv4Address() | IPv6Address():
            val = int(value)
            mask = all_ones(bitwidth)
        case _:
            raise _invalid_err("ternary", bitwidth, value)

    # TODO: Masked bits in value must be zero.
    return (encode_exact(val, bitwidth), encode_exact(mask, bitwidth))


def decode_ternary(
    data: bytes,
    mask: bytes,
    bitwidth: int,
    hint: DecodeFormat = DecodeFormat.DEFAULT,
) -> _TernaryReturn:
    "Decode a P4R Ternary value."
    if hint & DecodeFormat.STRING:
        # When decoding to string, check if mask can be represented
        # using LPM syntax or if the mask is all ones.
        prefix = mask_to_prefix(int.from_bytes(mask, "big"), bitwidth)
        if prefix >= 0:
            value = decode_exact(data, bitwidth, hint)
            if prefix == bitwidth:
                return value
            return f"{value}/{prefix}"

        tup = (
            decode_exact(data, bitwidth, hint),
            decode_exact(mask, bitwidth, hint),
        )
        return f"{tup[0]}/&{tup[1]}"

    return (decode_exact(data, bitwidth, hint), decode_exact(mask, bitwidth, hint))


def format_ternary(value: _TernaryValue, bitwidth: int, format: DecodeFormat) -> str:
    "Format a value as a string."
    data = encode_ternary(value, bitwidth)
    result = decode_ternary(data[0], data[1], bitwidth, format | DecodeFormat.STRING)
    assert isinstance(result, str)
    return result


# Range Values
# ~~~~~~~~~~~~
#
# Supported input values for `encode_range`:
#
# - str
#   o "lo...hi"
# - tuple[lo, hi]
#   o tuple[int, int]
#   o tuple[str, str]
#   o tuple[IPv4Address, IPv4Address] (bitwidth=32 only)
#   o tuple[IPv6Address, IPv6Address] (bitwidth=128 only)
#   o tuple[MACAddress, MACAddress]  (bitwidth=48 only)
#
# Supported output values for `decode_range`:
#
# - DEFAULT:
#   o tuple[int, int]
# - STRING:
#   o "lo...hi"
# - ADDRESS:
#   o See DEFAULT output.
# - ADDRESS, STRING
#   o See STRING output.

_RangeValue = str | tuple[_ExactValue, _ExactValue]
_RangeReturn = str | tuple[_ExactReturn, _ExactReturn]


def encode_range(value: _RangeValue, bitwidth: int) -> tuple[bytes, bytes]:
    "Encode a P4R Range value."
    assert value is not None

    match value:
        case str(val):
            low, high = val.split("...", 1)
        case (low, high):
            pass
        case _:  # pyright: ignore[reportUnnecessaryComparison]
            raise _invalid_err("range", bitwidth, value)

    return (encode_exact(low, bitwidth), encode_exact(high, bitwidth))


def decode_range(
    low: bytes,
    high: bytes,
    bitwidth: int,
    hint: DecodeFormat = DecodeFormat.DEFAULT,
) -> _RangeReturn:
    "Decode a P4R Range value."
    result = (decode_exact(low, bitwidth, hint), decode_exact(high, bitwidth, hint))

    match result[0]:
        case str():
            return f"{result[0]}...{result[1]}"
        case _:
            return result


def format_range(value: _RangeValue, bitwidth: int, format: DecodeFormat) -> str:
    "Format a value as a string."
    data = encode_range(value, bitwidth)
    result = decode_range(data[0], data[1], bitwidth, format | DecodeFormat.STRING)
    assert isinstance(result, str)
    return result


def encode_signed(value: int, bitwidth: int) -> bytes:
    "Encode a signed value."
    max_value = (1 << (bitwidth - 1)) - 1
    if value > max_value or value < -max_value - 1:
        raise _invalid_err("signed", bitwidth, value)

    size = p4r_minimum_string_size(bitwidth)
    return p4r_truncate_signed(value.to_bytes(size, "big", signed=True))


def decode_signed(data: bytes, _bitwidth: int) -> int:
    "Decode a signed value."
    return int.from_bytes(data, "big", signed=True)


def encode_enum(value: int, bitwidth: int) -> bytes:
    """Encode a P4R enum member into byte encoding."""
    if value >= (1 << bitwidth) or value < 0:
        raise _invalid_err("enum", bitwidth, value)

    size = p4r_minimum_string_size(bitwidth)
    return p4r_truncate(value.to_bytes(size, "big"))


def decode_enum(data: bytes, bitwidth: int) -> int:
    """Decode a P4R enum member as an unsigned integer."""
    if not data or bitwidth <= 0:
        raise _invalid_err("enum", bitwidth, data)

    value = int.from_bytes(data, "big")
    if value >= (1 << bitwidth):
        raise _invalid_err("enum", bitwidth, data)

    return value



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p4/config/v1/p4types.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ap4/config/v1/p4types.proto\x12\x0cp4.config.v1\"\xa1\x07\n\nP4TypeInfo\x12\x36\n\x07structs\x18\x01 \x03(\x0b\x32%.p4.config.v1.P4TypeInfo.StructsEntry\x12\x36\n\x07headers\x18\x02 \x03(\x0b\x32%.p4.config.v1.P4TypeInfo.HeadersEntry\x12\x41\n\rheader_unions\x18\x03 \x03(\x0b\x32*.p4.config.v1.P4TypeInfo.HeaderUnionsEntry\x12\x32\n\x05\x65nums\x18\x04 \x03(\x0b\x32#.p4.config.v1.P4TypeInfo.EnumsEntry\x12,\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x1d.p4.config.v1.P4ErrorTypeSpec\x12K\n\x12serializable_enums\x18\x06 \x03(\x0b\x32/.p4.config.v1.P4TypeInfo.SerializableEnumsEntry\x12\x39\n\tnew_types\x18\x07 \x03(\x0b\x32&.p4.config.v1.P4TypeInfo.NewTypesEntry\x1aN\n\x0cStructsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.p4.config.v1.P4StructTypeSpec:\x02\x38\x01\x1aN\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.p4.config.v1.P4HeaderTypeSpec:\x02\x38\x01\x1aX\n\x11HeaderUnionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.p4.config.v1.P4HeaderUnionTypeSpec:\x02\x38\x01\x1aJ\n\nEnumsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4EnumTypeSpec:\x02\x38\x01\x1a\x62\n\x16SerializableEnumsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32(.p4.config.v1.P4SerializableEnumTypeSpec:\x02\x38\x01\x1aL\n\rNewTypesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.p4.config.v1.P4NewTypeSpec:\x02\x38\x01\"\x83\x05\n\x0eP4DataTypeSpec\x12:\n\tbitstring\x18\x01 \x01(\x0b\x32%.p4.config.v1.P4BitstringLikeTypeSpecH\x00\x12(\n\x04\x62ool\x18\x02 \x01(\x0b\x32\x18.p4.config.v1.P4BoolTypeH\x00\x12.\n\x05tuple\x18\x03 \x01(\x0b\x32\x1d.p4.config.v1.P4TupleTypeSpecH\x00\x12+\n\x06struct\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x12+\n\x06header\x18\x05 \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x12\x31\n\x0cheader_union\x18\x06 \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x12;\n\x0cheader_stack\x18\x07 \x01(\x0b\x32#.p4.config.v1.P4HeaderStackTypeSpecH\x00\x12\x46\n\x12header_union_stack\x18\x08 \x01(\x0b\x32(.p4.config.v1.P4HeaderUnionStackTypeSpecH\x00\x12)\n\x04\x65num\x18\t \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x12*\n\x05\x65rror\x18\n \x01(\x0b\x32\x19.p4.config.v1.P4ErrorTypeH\x00\x12\x36\n\x11serializable_enum\x18\x0b \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x12-\n\x08new_type\x18\x0c \x01(\x0b\x32\x19.p4.config.v1.P4NamedTypeH\x00\x42\x0b\n\ttype_spec\"\x1b\n\x0bP4NamedType\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x0c\n\nP4BoolType\"\r\n\x0bP4ErrorType\"\xc5\x02\n\x17P4BitstringLikeTypeSpec\x12*\n\x03\x62it\x18\x01 \x01(\x0b\x32\x1b.p4.config.v1.P4BitTypeSpecH\x00\x12*\n\x03int\x18\x02 \x01(\x0b\x32\x1b.p4.config.v1.P4IntTypeSpecH\x00\x12\x30\n\x06varbit\x18\x03 \x01(\x0b\x32\x1e.p4.config.v1.P4VarbitTypeSpecH\x00\x12\x13\n\x0b\x61nnotations\x18\x04 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x06 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotationB\x0b\n\ttype_spec\"!\n\rP4BitTypeSpec\x12\x10\n\x08\x62itwidth\x18\x01 \x01(\x05\"!\n\rP4IntTypeSpec\x12\x10\n\x08\x62itwidth\x18\x01 \x01(\x05\"(\n\x10P4VarbitTypeSpec\x12\x14\n\x0cmax_bitwidth\x18\x01 \x01(\x05\"@\n\x0fP4TupleTypeSpec\x12-\n\x07members\x18\x01 \x03(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpec\"\xa8\x02\n\x10P4StructTypeSpec\x12\x36\n\x07members\x18\x01 \x03(\x0b\x32%.p4.config.v1.P4StructTypeSpec.Member\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x03 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x1aG\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\ttype_spec\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpec\"\xb1\x02\n\x10P4HeaderTypeSpec\x12\x36\n\x07members\x18\x01 \x03(\x0b\x32%.p4.config.v1.P4HeaderTypeSpec.Member\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x03 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x1aP\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\ttype_spec\x18\x02 \x01(\x0b\x32%.p4.config.v1.P4BitstringLikeTypeSpec\"\xac\x02\n\x15P4HeaderUnionTypeSpec\x12;\n\x07members\x18\x01 \x03(\x0b\x32*.p4.config.v1.P4HeaderUnionTypeSpec.Member\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x03 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x1a\x41\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12)\n\x06header\x18\x02 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"P\n\x15P4HeaderStackTypeSpec\x12)\n\x06header\x18\x01 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x0c\n\x04size\x18\x02 \x01(\x05\"[\n\x1aP4HeaderUnionStackTypeSpec\x12/\n\x0cheader_union\x18\x01 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x0c\n\x04size\x18\x02 \x01(\x05\"D\n\x0cKeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.p4.config.v1.Expression\"@\n\x10KeyValuePairList\x12,\n\x08kv_pairs\x18\x01 \x03(\x0b\x32\x1a.p4.config.v1.KeyValuePair\"Z\n\nExpression\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x15\n\x0bint64_value\x18\x02 \x01(\x03H\x00\x12\x14\n\nbool_value\x18\x03 \x01(\x08H\x00\x42\x07\n\x05value\"?\n\x0e\x45xpressionList\x12-\n\x0b\x65xpressions\x18\x01 \x03(\x0b\x32\x18.p4.config.v1.Expression\"\xd4\x01\n\x14StructuredAnnotation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x37\n\x0f\x65xpression_list\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.ExpressionListH\x00\x12\x36\n\x0ckv_pair_list\x18\x03 \x01(\x0b\x32\x1e.p4.config.v1.KeyValuePairListH\x00\x12\x35\n\x0fsource_location\x18\x04 \x01(\x0b\x32\x1c.p4.config.v1.SourceLocationB\x06\n\x04\x62ody\"<\n\x0eSourceLocation\x12\x0c\n\x04\x66ile\x18\x01 \x01(\t\x12\x0c\n\x04line\x18\x02 \x01(\x05\x12\x0e\n\x06\x63olumn\x18\x03 \x01(\x05\"\x89\x03\n\x0eP4EnumTypeSpec\x12\x34\n\x07members\x18\x01 \x03(\x0b\x32#.p4.config.v1.P4EnumTypeSpec.Member\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x04 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x03 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x1a\xab\x01\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x04 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x03 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"\xe6\x03\n\x1aP4SerializableEnumTypeSpec\x12\x34\n\x0funderlying_type\x18\x01 \x01(\x0b\x32\x1b.p4.config.v1.P4BitTypeSpec\x12@\n\x07members\x18\x02 \x03(\x0b\x32/.p4.config.v1.P4SerializableEnumTypeSpec.Member\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\x1a\xba\x01\n\x06Member\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"\"\n\x0fP4ErrorTypeSpec\x12\x0f\n\x07members\x18\x01 \x03(\t\"\x98\x01\n\x14P4NewTypeTranslation\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x16\n\x0csdn_bitwidth\x18\x02 \x01(\x05H\x00\x12\x42\n\nsdn_string\x18\x03 \x01(\x0b\x32,.p4.config.v1.P4NewTypeTranslation.SdnStringH\x00\x1a\x0b\n\tSdnStringB\n\n\x08sdn_type\"\xac\x02\n\rP4NewTypeSpec\x12\x35\n\roriginal_type\x18\x01 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpecH\x00\x12=\n\x0ftranslated_type\x18\x02 \x01(\x0b\x32\".p4.config.v1.P4NewTypeTranslationH\x00\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotationB\x10\n\x0erepresentationB-Z+github.com/p4lang/p4runtime/go/p4/config/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'p4.config.v1.p4types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z+github.com/p4lang/p4runtime/go/p4/config/v1'
  _globals['_P4TYPEINFO_STRUCTSENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_STRUCTSENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO_HEADERSENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO_HEADERUNIONSENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_HEADERUNIONSENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO_ENUMSENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_ENUMSENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO_SERIALIZABLEENUMSENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_SERIALIZABLEENUMSENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO_NEWTYPESENTRY']._loaded_options = None
  _globals['_P4TYPEINFO_NEWTYPESENTRY']._serialized_options = b'8\001'
  _globals['_P4TYPEINFO']._serialized_start=45
  _globals['_P4TYPEINFO']._serialized_end=974
  _globals['_P4TYPEINFO_STRUCTSENTRY']._serialized_start=472
  _globals['_P4TYPEINFO_STRUCTSENTRY']._serialized_end=550
  _globals['_P4TYPEINFO_HEADERSENTRY']._serialized_start=552
  _globals['_P4TYPEINFO_HEADERSENTRY']._serialized_end=630
  _globals['_P4TYPEINFO_HEADERUNIONSENTRY']._serialized_start=632
  _globals['_P4TYPEINFO_HEADERUNIONSENTRY']._serialized_end=720
  _globals['_P4TYPEINFO_ENUMSENTRY']._serialized_start=722
  _globals['_P4TYPEINFO_ENUMSENTRY']._serialized_end=796
  _globals['_P4TYPEINFO_SERIALIZABLEENUMSENTRY']._serialized_start=798
  _globals['_P4TYPEINFO_SERIALIZABLEENUMSENTRY']._serialized_end=896
  _globals['_P4TYPEINFO_NEWTYPESENTRY']._serialized_start=898
  _globals['_P4TYPEINFO_NEWTYPESENTRY']._serialized_end=974
  _globals['_P4DATATYPESPEC']._serialized_start=977
  _globals['_P4DATATYPESPEC']._serialized_end=1620
  _globals['_P4NAMEDTYPE']._serialized_start=1622
  _globals['_P4NAMEDTYPE']._serialized_end=1649
  _globals['_P4BOOLTYPE']._serialized_start=1651
  _globals['_P4BOOLTYPE']._serialized_end=1663
  _globals['_P4ERRORTYPE']._serialized_start=1665
  _globals['_P4ERRORTYPE']._serialized_end=1678
  _globals['_P4BITSTRINGLIKETYPESPEC']._serialized_start=1681
  _globals['_P4BITSTRINGLIKETYPESPEC']._serialized_end=2006
  _globals['_P4BITTYPESPEC']._serialized_start=2008
  _globals['_P4BITTYPESPEC']._serialized_end=2041
  _globals['_P4INTTYPESPEC']._serialized_start=2043
  _globals['_P4INTTYPESPEC']._serialized_end=2076
  _globals['_P4VARBITTYPESPEC']._serialized_start=2078
  _globals['_P4VARBITTYPESPEC']._serialized_end=2118
  _globals['_P4TUPLETYPESPEC']._serialized_start=2120
  _globals['_P4TUPLETYPESPEC']._serialized_end=2184
  _globals['_P4STRUCTTYPESPEC']._serialized_start=2187
  _globals['_P4STRUCTTYPESPEC']._serialized_end=2483
  _globals['_P4STRUCTTYPESPEC_MEMBER']._serialized_start=2412
  _globals['_P4STRUCTTYPESPEC_MEMBER']._serialized_end=2483
  _globals['_P4HEADERTYPESPEC']._serialized_start=2486
  _globals['_P4HEADERTYPESPEC']._serialized_end=2791
  _globals['_P4HEADERTYPESPEC_MEMBER']._serialized_start=2711
  _globals['_P4HEADERTYPESPEC_MEMBER']._serialized_end=2791
  _globals['_P4HEADERUNIONTYPESPEC']._serialized_start=2794
  _globals['_P4HEADERUNIONTYPESPEC']._serialized_end=3094
  _globals['_P4HEADERUNIONTYPESPEC_MEMBER']._serialized_start=3029
  _globals['_P4HEADERUNIONTYPESPEC_MEMBER']._serialized_end=3094
  _globals['_P4HEADERSTACKTYPESPEC']._serialized_start=3096
  _globals['_P4HEADERSTACKTYPESPEC']._serialized_end=3176
  _globals['_P4HEADERUNIONSTACKTYPESPEC']._serialized_start=3178
  _globals['_P4HEADERUNIONSTACKTYPESPEC']._serialized_end=3269
  _globals['_KEYVALUEPAIR']._serialized_start=3271
  _globals['_KEYVALUEPAIR']._serialized_end=3339
  _globals['_KEYVALUEPAIRLIST']._serialized_start=3341
  _globals['_KEYVALUEPAIRLIST']._serialized_end=3405
  _globals['_EXPRESSION']._serialized_start=3407
  _globals['_EXPRESSION']._serialized_end=3497
  _globals['_EXPRESSIONLIST']._serialized_start=3499
  _globals['_EXPRESSIONLIST']._serialized_end=3562
  _globals['_STRUCTUREDANNOTATION']._serialized_start=3565
  _globals['_STRUCTUREDANNOTATION']._serialized_end=3777
  _globals['_SOURCELOCATION']._serialized_start=3779
  _globals['_SOURCELOCATION']._serialized_end=3839
  _globals['_P4ENUMTYPESPEC']._serialized_start=3842
  _globals['_P4ENUMTYPESPEC']._serialized_end=4235
  _globals['_P4ENUMTYPESPEC_MEMBER']._serialized_start=4064
  _globals['_P4ENUMTYPESPEC_MEMBER']._serialized_end=4235
  _globals['_P4SERIALIZABLEENUMTYPESPEC']._serialized_start=4238
  _globals['_P4SERIALIZABLEENUMTYPESPEC']._serialized_end=4724
  _globals['_P4SERIALIZABLEENUMTYPESPEC_MEMBER']._serialized_start=4538
  _globals['_P4SERIALIZABLEENUMTYPESPEC_MEMBER']._serialized_end=4724
  _globals['_P4ERRORTYPESPEC']._serialized_start=4726
  _globals['_P4ERRORTYPESPEC']._serialized_end=4760
  _globals['_P4NEWTYPETRANSLATION']._serialized_start=4763
  _globals['_P4NEWTYPETRANSLATION']._serialized_end=4915
  _globals['_P4NEWTYPETRANSLATION_SDNSTRING']._serialized_start=4892
  _globals['_P4NEWTYPETRANSLATION_SDNSTRING']._serialized_end=4903
  _globals['_P4NEWTYPESPEC']._serialized_start=4918
  _globals['_P4NEWTYPESPEC']._serialized_end=5218
# @@protoc_insertion_point(module_scope)
